@@ -2,8 +2,8 @@
   This file is part of usb_modeswitch, a mode switching tool for controlling
   the mode of 'multi-state' USB devices
 
-  Version 2.2.0, 2014/05/29
-  Copyright (C) 2007 - 2014  Josua Dietze
+  Version 2.2.5, 2015/07/16
+  Copyright (C) 2007 - 2015  Josua Dietze
 
   Config file parsing stuff borrowed from Guillaume Dargaud
   (http://www.gdargaud.net/Hack/SourceCode.html)
@@ -60,6 +60,8 @@ int hex2byte(const char *hex);
 int hexstr2bin(const char *hex, char *buffer, int len);
 void printVersion();
 void printHelp();
+void close_all();
+void abort();
 int readArguments(int argc, char **argv);
 void deviceDescription();
 int deviceInquire();
